#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <math.h>

#include "volume.h"
#include "utils.h"

typedef struct {
  int next;
  volume_item_t *v; /* vol: -200=empty
                     *      [-1000-60..-1000+60]=mute (and keep old value)
                     *      [-60..60]=volume in dB */
  int vsize;
  int vmaxsize;
  pthread_mutex_t lock;
  pthread_cond_t cond;
  int dirty;
  int client_connected;  /* only 1 client at a time */
} volume_t;

static void lock(volume_t *v) { if (pthread_mutex_lock(&v->lock)) abort(); }
static void unlock(volume_t *v) {if(pthread_mutex_unlock(&v->lock))abort();}

static void csignal(volume_t *v)
{
  if (pthread_cond_broadcast(&v->cond)) abort();
}

static void cwait(volume_t *v)
{
  if (pthread_cond_wait(&v->cond, &v->lock)) abort();
}

void *create_volume_module(void)
{
  volume_t *v = calloc(1, sizeof(*v)); if (v == NULL) abort();
  if (pthread_mutex_init(&v->lock, NULL)) abort();
  if (pthread_cond_init(&v->cond, NULL)) abort();
  return v;
}

void delete_volume_module(void *_)
{
  volume_t *m = _;
  free(m->v);
  free(m);
}

int new_volume(void *_)
{
  volume_t *m = _;
  int i;
  int ret;

  lock(m);

  if (m->vsize == m->vmaxsize) {
    int inc = 256;
    m->vmaxsize += inc;
    m->v = realloc(m->v, m->vmaxsize * sizeof(*m->v)); if(m->v==NULL)abort();
    for (i = m->vmaxsize - inc; i < m->vmaxsize; i++) m->v[i].vol = -200;
  }

  for (i = 0; i < m->vmaxsize; i++) if (m->v[i].vol == -200) break;
  m->v[i].vol = 0;
  m->v[i].id = ret = m->next;
  m->vsize++;
  m->next++;

  m->dirty = 1;
  csignal(m);

  unlock(m);

  return ret;
}

void free_volume(void *_, int id)
{
  volume_t *m = _;
  int i;

  lock(m);

  for (i = 0; i < m->vmaxsize; i++)
    if (m->v[i].vol != -200 && m->v[i].id == id) break;
  if (i == m->vmaxsize)
    {printf("error: free_volume: id not found %d, do nothing\n",id);goto end;}
  m->v[i].vol = -200;
  m->v[i].id = -1;
  m->vsize--;

  m->dirty = 1;
  csignal(m);

end:
  unlock(m);
}

int get_volume(void *_, int id)
{
  volume_t *m = _;
  int i;
  int ret = 0;

  lock(m);

  for (i = 0; i < m->vmaxsize; i++)
    if (m->v[i].vol != -200 && m->v[i].id == id) break;
  if (i == m->vmaxsize)
    {printf("error: get_volume: id not found %d, return 0\n",id);goto end;}
  ret = m->v[i].vol;

end:
  unlock(m);

  return ret;
}

void set_volume(void *_, int id, int vol)
{
printf("set volume id %d vol %d\n", id, vol);
  volume_t *m = _;
  int i;

  if (!((vol >= -1000-60 && vol <= -1000+60) || (vol >= -60 && vol <= 60)))
    { printf("error: set_volume: bad volume %d, do nothing\n", vol); return; }

  lock(m);

  for (i = 0; i < m->vmaxsize; i++)
    if (m->v[i].vol != -200 && m->v[i].id == id) break;
  if (i == m->vmaxsize)
    {printf("error: set_volume: id not found %d, do nothing\n",id);goto end;}
  m->v[i].vol = vol;

  m->dirty = 1;
  csignal(m);

end:
  unlock(m);
}

float volume_db2lin(int vol)
{
  if (vol < -100) return 0;
  return pow(10, vol/10.);
}

typedef struct {
  int fd;
  volatile int alive;
  volume_t *m;
} sender_data_t;

static int write_int(int fd, int v)
{
  char b[4];
  b[0] = v & 255;
  b[1] = (v >>  8) & 255;
  b[2] = (v >> 16) & 255;
  b[3] = (v >> 24) & 255;
  int ret = fullwrite(fd, b, 4) != 4 ? -1 : 0;
  return ret;
}

static int read_int(int fd, int *v)
{
  char b[4];
  int ret = fullread(fd, b, 4);
  if (ret != 4) return -1;

  *v = ( b[0] & 255) |
       ((b[1] & 255) <<  8) |
       ((b[2] & 255) << 16) |
       ((b[3] & 255) << 24);
  return 0;
}

/* called with lock */
static int state_to_client(int fd, volume_t *m)
{
  int i;
  int count = 0;

  for (i = 0; i < m->vmaxsize; i++)
    if (m->v[i].vol != -200) count++;

  if (write_int(fd, count) == -1) return -1;

  for (i = 0; i < m->vmaxsize; i++)
    if (m->v[i].vol != -200)
      if (write_int(fd, m->v[i].id) == -1 ||
          write_int(fd, m->v[i].vol) == -1) return -1;

  return 0;
}

/* called without lock */
static int state_from_client(int fd, volume_t *m)
{
  int i;
  int count;
  int id;
  int vol;

  if (read_int(fd, &count) == -1) return -1;

  for (i = 0; i < count; i++) {
    if (read_int(fd, &id) == -1 ||
        read_int(fd, &vol) == -1) return -1;
    set_volume(m, id, vol);
  }

  return 0;
}

static void *sender_thread(void *_)
{
  sender_data_t *s = _;
  volume_t *m = s->m;

  printf("volume:sender_thread starts\n");

  lock(m);

more:
  while (s->alive && !m->dirty) cwait(m);
  if (!s->alive) goto end;

  if (state_to_client(s->fd, m) == -1) goto end;
  m->dirty = 0;
  goto more;

end:
  s->alive = 2;
  csignal(m);

  unlock(m);

  printf("volume:sender_thread ends\n");

  return 0;
}

void run_volume_client(void *_, int fd)
{
  sender_data_t d;
  volume_t *m = _;
  lock(m);
  if (m->client_connected) { unlock(m); return; }
  m->client_connected = 1;
  m->dirty = 1;
  unlock(m);

  d.fd = fd;
  d.alive = 1;
  d.m = m;
  new_thread(sender_thread, &d);

  while (1)
    if (state_from_client(fd, m) == -1) break;

  lock(m);

  if (d.alive != 2) {
    d.alive = 0;
    csignal(m);
  }

  while (d.alive != 2) cwait(m);

  m->client_connected = 0;
  unlock(m);

  printf("run_volume_client ends\n");
}
